/**
 * @file drslot_chrp_cpu.c
 *
 *
 * Copyright (C) IBM Corporation 2006
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <librtas.h>
#include "dr.h"
#include "drcpu.h"
#include "drpci.h"
#include "ofdt.h"

struct cpu_operation;
typedef int (cpu_op_func_t) (struct options *);

struct parm_to_func {
	char	*parmname;
	cpu_op_func_t *func;
};


static char *usagestr = "-c cpu {-a | -r} {-q <quantity> -p {variable_weight | ent_capacity} [-s drc_names]";

/**
 * cpu_usage
 *
 */
void
cpu_usage(char **pusage)
{
	*pusage = usagestr;
}
/**
 * add_cpus
 *
 * Attempt to acquire and online the given number of cpus.
 * This function calls itself recursively to simplify recovery
 * actions in case of an error.  This is intended only for the case
 * where the user does not specify a drc-name.
 *
 * The final steps are to display the drc-names value to stdout and
 * return with 0.
 *
 * @param nr_cpus
 * @returns 0 on success, !0 otherwise
 */
static int
add_cpus(int nr_cpus, struct dr_info *dr_info)
{
	int rc = 0;
	struct cpu_info *cpu;
	char *drc_name;

	cpu = acquire_cpu(0, dr_info);

	if (cpu == NULL) {
		err_msg("Unable to acquire additional CPU\n");
		return 1;
	}

	drc_name = cpu_drc_name(cpu, dr_info);

	rc = set_cpu_state(cpu, 1, dr_info);

	if (rc) {
		/* Roll back the operation.  Is this the correct behavior? */
		err_msg("Unable to online %s\n", drc_name);
		set_cpu_state(cpu, 0, dr_info);
		release_cpu(cpu, dr_info);
		return 1;
	}

	if (nr_cpus > 1)
		rc = add_cpus(nr_cpus - 1, dr_info);

	if (rc) {
		/* Roll back the operation.  Is this the correct behavior? */
		release_cpu(cpu, dr_info);
		return 1;
	}

	fprintf(stdout, "%s\n", drc_name);

	return 0;
}

/**
 * remove_cpus
 *
 * Attempt to offline and release to the hypervisor the given number of
 * cpus.  This functions calls itself recursively to simplify recovery
 * actions in the case of an error.  This is intended only for the case
 * where the user does not specify a drc-name.
 *
 * From "Design Specification for AIX Configuration Support of
 * Dynamic Reconfiguration including the drmgr command and drslot for
 * memory, processors, and PCI slots" Version 1.2:
 *
 * Section V. Part B. Item 2. "drslot_chrp_cpu -r -c cpu"
 * "Once the resource has been released by the kernel, all the following
 * steps are taken.  Errors are ignored.  The code continues releasing the
 * resource by using RTAS services. ..."
 * "If successful, the code displays the drc-names value to stdout and
 * returns with 0, else displays an error message to stderr and returns with
 * non-zero."
 *
 * @param nr_cpus
 * @returns 0 on success, !0 otherwise
 */
static int
remove_cpus(int nr_cpus, struct dr_info *dr_info)
{
	int rc = 0, i = 0;
	struct thread_info *thread = NULL;
	struct cpu_info *cpu;
	char *drc_name;

	/* Find the first online thread and kill its cpu. */
	while ((thread = id_to_thread(i++, dr_info))) {
		if (get_thread_state(thread) == 1)
			break;
	}

	cpu = thread_to_cpu(thread, dr_info);
	drc_name = cpu_drc_name(cpu, dr_info);

	rc = set_cpu_state(cpu, 0, dr_info);

	if (rc) {
		/* Need to online any threads we took offline */
		set_cpu_state(cpu, 1, dr_info);
		err_msg("Could not offline %s\n", drc_name);
		return rc;
	}

	/* cpu is invalid after release_cpu, so no recovery
	 * steps seem feasible.  We could copy the cpu name
	 * and look it up again if the operation fails.
	 */
	rc = release_cpu(cpu, dr_info);
	if (rc) {
		err_msg("Offlined but could not release %s\n", drc_name);
		return rc;
	}

	fprintf(stdout, "%s\n", drc_name);
	if (nr_cpus > 1)
		rc = remove_cpus(nr_cpus - 1, dr_info);

	return rc;
}

/**
 * default_cpu_func
 * @brief Add or remove some physical cpus
 *
 * @param opts
 * @returns 0 on success, !0 otherwise
 */
static int
default_cpu_func(struct options *opts)
{
	struct dr_info dr_info;
	int lock_id;
	int rc = -1;

	if (! cpu_dlpar_capable()) {
		dbg("Machine sis not DLPAR capable.\n");
		return -1;
	}

	lock_id = dr_lock(opts->timeout);

	if (init_cpu_drc_info(&dr_info))
		goto default_cpu_func_exit;

	/* If the user specified a drc-name, we handle that case
	 * specially here.
	 */
	if (opts->usr_drc_name) {
		int index = drc_name_to_index(opts->usr_drc_name,
					      dr_info.all_drcs);

		if (!index) {
			err_msg("invalid drc-name %s specified\n",
				opts->usr_drc_name);
			goto default_cpu_func_exit;
		}

		if (opts->action == ADD) {
			struct cpu_info *cpu;

			if ((cpu = acquire_cpu(index, &dr_info)))
				rc = set_cpu_state(cpu, 1, &dr_info);
			else
				err_msg("Could not acquire %s\n",
					opts->usr_drc_name);

			if (rc != 0) {
				err_msg("Could not online %s\n",
					opts->usr_drc_name);
				set_cpu_state(cpu, 0, &dr_info);
				release_cpu(cpu, &dr_info);
				goto default_cpu_func_exit;
			}

			fprintf(stdout, "%s\n", opts->usr_drc_name);

		} else if (opts->action == REMOVE) {
			struct cpu_info *cpu;

			cpu = drc_index_to_cpu(index, &dr_info);
			if (cpu == NULL) {
				err_msg("no active cpu for drc-name %s\n",
					opts->usr_drc_name);
				goto default_cpu_func_exit;
			}

			if ((rc = set_cpu_state(cpu, 0, &dr_info)) == 0)
				rc = release_cpu(cpu, &dr_info);
			else
				err_msg("Could not offline %s\n",
					opts->usr_drc_name);

			if (rc != 0) /* warn */
				err_msg("Offlined but could not release %s\n",
					opts->usr_drc_name);
			else
				fprintf(stdout, "%s\n", opts->usr_drc_name);
		}

	} else { /* no drc name specified */
		if (opts->action == ADD)
			rc = add_cpus(opts->quantity, &dr_info);
		if (opts->action == REMOVE)
			rc = remove_cpus(opts->quantity, &dr_info);
	}

default_cpu_func_exit:
	free_cpu_drc_info(&dr_info);
	dr_unlock(lock_id);

	return rc;
}

/**
 * smt_thread_func
 * @brief Act upon logical cpus/threads
 *
 * @param opts
 * @returns 0 on success, !0 otherwise
 */
static int
smt_threads_func(struct options *opts)
{
	struct dr_info dr_info;
	int lock_id;
	int rc = -1;

	if (opts->quantity != 1) {
		err_msg("Quantity (-q) may not be specified with"
			" the smt_threads parameter\n");
		return -1;
	}

	lock_id = dr_lock(opts->timeout);

	if (init_cpu_drc_info(&dr_info))
		goto smt_threads_func_exit;


	if (! smt_enabled(&dr_info)) {
		err_msg("SMT functions not available on this system.\n");
		goto smt_threads_func_exit;
	}

	if (opts->usr_drc_name) {
		int index = drc_name_to_index(opts->usr_drc_name,
					      dr_info.all_drcs);
		struct cpu_info *cpu;

		if (! index) {
			err_msg("invalid drc-name %s specified\n",
				opts->usr_drc_name);
			goto smt_threads_func_exit;
		}

		cpu = drc_index_to_cpu(index, &dr_info);
		if (cpu == NULL) {
			err_msg("no active cpu for drc-name %s",
				opts->usr_drc_name);
			goto smt_threads_func_exit;
		}

		if (opts->action == ADD)
			rc = cpu_enable_smt(cpu, &dr_info);
		else if (opts->action == REMOVE)
			rc = cpu_disable_smt(cpu, &dr_info);

	} else { /* no drc name given, action is system-wide */
		if (opts->action == ADD)
			rc = system_enable_smt(&dr_info);
		if (opts->action == REMOVE)
			rc = system_disable_smt(&dr_info);
	}

smt_threads_func_exit:
	free_cpu_drc_info(&dr_info);
	dr_unlock(lock_id);

	return rc;
}

/* Map parameter names to functions. */
static struct parm_to_func parm_func_table[] =
{
	{
		.parmname = "smt_threads",
		.func = smt_threads_func
	},
	{
		.parmname = "ent_capacity",
		.func = update_sysparm
	},
	{
		.parmname = "variable_weight",
		.func = update_sysparm
	}
};

/**
 * lookup_cpu_func
 *
 * @param name
 * @returns pointer to cpu function on success, NULL otherwise
 */
static cpu_op_func_t *
lookup_cpu_func(const char *name)
{
	int i = sizeof(parm_func_table) / sizeof(parm_func_table[0]);

	if (name == NULL)
		return default_cpu_func;

	while (--i >= 0)
		if (! strcmp(parm_func_table[i].parmname, name))
			return parm_func_table[i].func;

	return NULL;
}

int
valid_cpu_options(struct options *opts)
{
	/* default to a quantity of 1 */
	if ((opts->quantity == 0) && (opts->usr_drc_name == NULL))
		opts->quantity = 1;

	if ((opts->action != ADD) && (opts->action != REMOVE)) {
		err_msg("Invalid action specified\n");
		return -1;
	}

	return 0;
}

int
drslot_chrp_cpu(struct options *opts)
{
	cpu_op_func_t *func;

	func = lookup_cpu_func(opts->p_option);
	if (func == NULL) {
		err_msg("Invalid parameter -p %s\n", opts->p_option);
		return -1;
	}

	return func(opts);
}
